"""
GOBNILP Solver for Exact Bayesian network
Structure Learning with Integer Linear
Optimization.

This code relies on the "pyGOBN" package, which
implements a Python wrapper for the GOBNILP
project - which itself builds on the SCIP
Optimization engine for building Integer Linear
Programs to solve the BN structure learning problem
EXACTLY.

Th pyGOBN module - which must be installed from source - allows 
users to create/alter a vast number of global parameter settings 
(i.e. wall-time, parent limits, sparsity, equivalent 
sample size, etc), along with various network constraints 
(i.e. required edges, disallowed edges, and independence
requirements).

NOTE: You must download and install pyGOBN to use this functionality.
Visit github.com/ncullen93/pyGOBN to get the source code, then run
'python setup.py install' in the main pyGOBN directory to install the
module for your local python distribution.

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

def ilp(data, settings=None, edge_reqs=None, nonedge_reqs=None, ind_reqs=None):
	"""
	Learning the EXACT optimal Bayesian network structure from data using
	Integer Linear Programming from the GOBNILP project.
	"""
	try:
		from pyGOBN import GOBN
	except ImportError:
		print('You must download & install pyGOBN to use this functionality.\
		 Visit github.com/ncullen93/pyGOBN')

	gobn = GOBN()
	gobn.set_settings(settings)
	gobn.set_constraints(edge_reqs=edge_reqs, nonedge_reqs=nonedge_reqs, ind_reqs=ind_reqs)



"""
**************
Markov Blanket
Unit Test
**************
"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np
import pandas as pd

from external.pyBN.readwrite.read import read_bn
from external.pyBN.independence.markov_blanket import markov_blanket

class ConstraintTestsTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn = read_bn(os.path.join(self.dpath,'cmu.bn'))

	def tearDown(self):
		pass

	def test_markov_blanket(self):
		self.assertDictEqual(markov_blanket(self.bn),
			{'Alarm': ['Earthquake', 'Burglary', 'JohnCalls', 'MaryCalls'],
			 'Burglary': ['Alarm', 'Earthquake'],
			 'Earthquake': ['Alarm', 'Burglary'],
			 'JohnCalls': ['Alarm'],
			 'MaryCalls': ['Alarm']})


"""
******************
Orient Edges from
Structure Learning
******************

[1] Chickering. "Learning Equivalence Classes of Bayesian
Network Structues"
http://www.jmlr.org/papers/volume2/chickering02a/chickering02a.pdf

[2] Pellet and Elisseff, "Using Markov Blankets for Causal Structure Learning"
"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.utils.independence_tests import mi_test

import itertools
from copy import copy

def orient_edges_MB(edge_dict, Mb, data, alpha):
	"""
	Orient edges from a Markov Blanket based on the rules presented
	in Margaritis' Thesis pg. 35. This method is used
	for structure learning algorithms that return/resolve
	a markov blanket - i.e. growshrink and iamb.

	Also, see [2] for good full pseudocode.

	# if there exists a variable Z in N(X)-N(Y)-{Y}
	# such that Y and Z are dependent given S+{X} for
	# all S subset of T, where
	# T is smaller of B(Y)-{X,Z} and B(Z)-{X,Y}

	Arguments
	---------
	*edge_dict* : a dictionary, where
		key = node and value = list
		of neighbors for key. Note: there
		MUST BE duplicates in edge_dict ->
		i.e. each edge should be in edge_dict
		twice since Y in edge_dict[X] and
		X in edge_dict[Y]

	*blanket* : a dictionary, where
		key = node and value = list of
		nodes in the markov blanket of node

	*data* : a nested numpy array

	*alpha* : a float
		Probability of Type II error.

	Returns
	-------
	*d_edge_dict* : a dictionary
		Dictionary of directed edges, so
		there are no duplicates

	Effects
	-------
	None

	Notes
	-----

	"""
	for X in edge_dict.keys():
		for Y in edge_dict[X]:
			nxy = set(edge_dict[X]) - set(edge_dict[Y]) - {Y}

			for Z in nxy:
				by = set(Mb[Y]) - {X} - {Z}
				bz = set(Mb[Z]) - {X} - {Y}
				T = min(by,bz)
				if len(T)>0:
					for i in range(len(T)):
						for S in itertools.combinations(T,i):
							cols = (Y,Z,X) + tuple(S)
							pval = mi_test(data[:,cols])
							if pval < alpha:
								if Y in edge_dict[X]:
									edge_dict[X].remove(Y)
							else:
								if Y in edge_dict[X]:
									edge_dict[Y].remove(X)
				else:
					cols = (Y,Z,X)
					pval = mi_test(data[:,cols])
					if pval < alpha:
						if Y in edge_dict[X]:
							edge_dict[X].remove(Y)
					else:
						if X in edge_dict[Y]:
							edge_dict[Y].remove(X)
	return edge_dict

def orient_edges_gs2(edge_dict, Mb, data, alpha):
	"""
	Similar algorithm as above, but slightly modified for speed?
	Need to test.
	"""
	d_edge_dict = dict([(rv,[]) for rv in edge_dict])
	for X in edge_dict.keys():
		for Y in edge_dict[X]:
			nxy = set(edge_dict[X]) - set(edge_dict[Y]) - {Y}
			for Z in nxy:
				if Y not in d_edge_dict[X]:
					d_edge_dict[X].append(Y) # SET Y -> X
				B = min(set(Mb[Y]) - {X} - {Z},set(Mb[Z]) - {X} - {Y})
				for i in range(len(B)):
					for S in itertools.combinations(B,i):
						cols = (Y,Z,X) + tuple(S)
						pval = mi_test(data[:,cols])
						if pval < alpha and X in d_edge_dict[Y]: # Y IS independent of Z given S+X
							d_edge_dict[Y].remove(X)
				if X in d_edge_dict[Y]:
					break
	return d_edge_dict

def orient_edges_CS(edge_dict, block_dict):
	"""
	Orient edges from Collider Sets. This is a little
	different than orienting edges from a markov blanket.
	The collider set-based algorithm is used for the 
	Path-Condition (PC) algorithm.

	See [2] for	good full pseudocode.

	The orientation step will proceed by looking
	for sets of three variables {X, Y,Z} such that
	edges X - Z, Y - Z are in the graph by not the
	edge X - Y . Then, if Z not in block_dict[x][y] , it orients the
	edges from X to Z and from Y to Z creating a
	v-structure: X -> Z <- Y

	Arguments
	---------
	*edge_dict* : a dictionary, where
		key = vertex and value = list of its neighbors.
		NOTE: this is undirected, so the edges are duplicated.

	*block_dict* : a dictionary, where
		key = X, value = another dictionary where
		key = Y, value = set S such that I(X,Y|S)

	Returns
	-------
	*d_edge_dict* : an directed version of original edge_dict

	Effects
	-------
	None

	Notes
	-----
	"""
	d_edge_dict = dict([(rv,[]) for rv in edge_dict.keys()])
	for x in edge_dict.keys():
		for z in edge_dict[x]:
			for y in edge_dict[z]:
				if y!=x and x not in edge_dict[y] and y not in edge_dict[x]:
					if y in block_dict[x]:
						if z not in block_dict[x][y]:
							if z not in d_edge_dict[x]:
								d_edge_dict[x].append(z)
							if z not in d_edge_dict[y]:
								d_edge_dict[y].append(z)
						else:
							if x not in d_edge_dict[z]:
								d_edge_dict[z].append(x)
							if y not in d_edge_dict[z]:
								d_edge_dict[z].append(y)

	return d_edge_dict

















# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
A module for creating and managing node data. Node data in this library can have many types, dependent on whether the conditional probability distributions are discrete, linear Gaussian, or hybrid, and on whether the Bayesian network is static or dynamic. For example input files, see :doc:`unittestdict`, :doc:`unittesthdict`, :doc:`unittestlgdict`, and :doc:`unittestdyndict`.

'''

from bamt.external.libpgm.dictionary import Dictionary


class NodeData(Dictionary):
    '''
    This class represents the node data for each node in a graph. If the Bayesian network is static, it contains the attribute *Vdata*. If the Bayesian network is dynamic, it contains two attributes, *initial_Vdata* and *twotbn_Vdata*. If the Bayesian network has hybrid CPDs, it contains the additional attribute *nodes*.
    
    '''
    def __init__(self):
        self.Vdata = None
        '''A dictionary of node data.'''
        self.initial_Vdata = None
        '''In dynamic graphs, a dictionary containing node data for the initial time interval.'''
        self.twotbn_Vdata = None
        '''In dynamic graphs, a dictionary containing node data for every time step interval after the first one.'''
        self.nodes = None
        '''In hybrid graphs, a dictionary of {key:value} pairs linking the name of each node (the key) to a clas instance (the value) which represents the node, its data, and its sampling function.'''


    def load(self, path):
        '''
        Load node data from an input file located at *path*. Input file must be a plaintext .txt file with a JSON-style representation of a dict. The dict must have the top-level key ``Vdata`` or two top-level keys, ``initial_Vdata`` and ``twotbn_Vdata``. For example:: 

            {
                "Vdata": {
                    "<vertex 1>": <dict containing vertex 1 data>,
                    ...
                    "<vertex n>": <dict containing vertex n data>
                }
            }

        or::

            {
                "initial_Vdata": {
                    "<vertex 1>": <dict containing vertex 1 data>,
                    ...
                    "<vertex n>": <dict containing vertex n data>
                }
                "twotbn_Vdata": {
                    "<vertex 1>": <dict containing vertex 1 data>,
                    ...
                    "<vertex n>": <dict containing vertex n data>
                }
            }
            
        The function takes the following arguments:
            1. *path* -- The path to the text file that contains input data (e.g., "mydictionary.txt")
        
        In the static case, it modifies *Vdata* to hold the dictionary found at path. In the dynamic case, it modifies the *initial_Vdata* and *twotbn_Vdata* attributes to hold the dictionaries found at path.
            
        '''
        self.dictload(path)

        # try to load both for normal and dynamic cases
        try: 
            self.Vdata = self.alldata["Vdata"]
        except KeyError:
            try:
                self.initial_Vdata = self.alldata["initial_Vdata"]
                self.twotbn_Vdata = self.alldata["twotbn_Vdata"]
            except KeyError:
                print ("Error: NodeData did not recognize input file format.")


        # free unused memory
        del self.alldata

    def entriestoinstances(self):
        '''
        For each node, convert dictionary entry to class instance.

        This method is used only when dealing with Hybrid Bayesian networks as found in the :doc:`hybayesiannetwork` module. 

        The type of the node must be located in the 'type' attribute of the node's dictionary entry. To see an example of such a dictionary, see :doc:`unittesthdict`. This type is used to instantiate a corresponding class from external.libpgm/CPDtypes/, and store the node's dictionary info in that class. Thus we lose none of the dictionary data, yet we gain the ability to use the instantiated class's built-in function to choose its own outcome based on the outcomes of its parents.

        In order for this method to be called, the self.Vdata attribute must have dictionary entries of the following form::

            <vertex name>: {
                'type': <type of node -- must correspond to module in /CPDtypes>,
                'parents': <array of parents of node>,
                'children': <array of children of node>,
                <whatever further entries are required by the type*> 
            }

        For instance, type "discrete" requires a "cprob" entry, while type "lg"
        requires "mean_base", "mean_scal", and "variance" entries.

        The function draws on the data in the *Vdata* attribute, and instantiates the attribute *nodes*, which is a dictionary of {name: instance} pairs where 'name' is the name of the node and 'instance' is a class instance containing the node data and the proper sampling function.

        '''
        # declare result dict
        rarray = dict()

        # transform into class instances
        for entry in self.Vdata.keys():
            
            # import module containing class
            path = str(self.Vdata[entry]["type"])
            exec("import external.libpgm." + path + " as " + path)

            # instantiate class
            exec("tmpnode = " + path + "." + str.capitalize(path) + "(self.Vdata[entry])")

            # append to array
            exec("rarray['" + str(entry) + "'] = tmpnode")

        self.nodes = rarray




"""
********
UnitTest
BayesNet
********

Method	Checks that
assertEqual(a, b)	a == b	 
assertNotEqual(a, b)	a != b	 
assertTrue(x)	bool(x) is True	 
assertFalse(x)	bool(x) is False	 
assertIs(a, b)	a is b
assertIsNot(a, b)	a is not b
assertIsNone(x)	x is None
assertIsNotNone(x)	x is not None
assertIn(a, b)	a in b
assertNotIn(a, b)	a not in b
assertIsInstance(a, b)	isinstance(a, b)
assertNotIsInstance(a, b)	not isinstance(a, b)

assertAlmostEqual(a, b)	round(a-b, 7) == 0	 
assertNotAlmostEqual(a, b)	round(a-b, 7) != 0	 
assertGreater(a, b)	a > b
assertGreaterEqual(a, b)	a >= b
assertLess(a, b)	a < b
assertLessEqual(a, b)	a <= b

assertListEqual(a, b)	lists
assertTupleEqual(a, b)	tuples
assertSetEqual(a, b)	sets or frozensets
assertDictEqual(a, b)	dicts

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
from external.pyBN.classes.bayesnet import BayesNet
from external.pyBN.readwrite.read import read_bn

import os
from os.path import dirname


class BayesNetTestCase(unittest.TestCase):

	def setUp(self):
		self.bn = BayesNet()
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn_bif = read_bn(os.path.join(self.dpath,'cancer.bif'))
		self.bn_bn = read_bn(os.path.join(self.dpath,'cmu.bn'))

	def tearDown(self):
		pass

	def test_isinstance(self):
		self.assertIsInstance(self.bn,BayesNet)

	def test_V_bif(self):
		self.assertListEqual(self.bn_bif.V,
			['Smoker', 'Pollution', 'Cancer', 'Xray', 'Dyspnoea'])

	def test_E_bif(self):
		self.assertDictEqual(self.bn_bif.E,
			{'Cancer': ['Xray', 'Dyspnoea'],
			 'Dyspnoea': [],
			 'Pollution': ['Cancer'],
			 'Smoker': ['Cancer'],
			 'Xray': []})

	def test_F_bif(self):
		self.assertDictEqual(self.bn_bif.F,
			{'Cancer': {'cpt': [0.03, 0.97, 0.05, 0.95, 0.001, 0.999, 0.02, 0.98],
				  'parents': ['Pollution', 'Smoker'],
				  'values': ['True', 'False']},
			 'Dyspnoea': {'cpt': [0.65, 0.35, 0.3, 0.7],
				  'parents': ['Cancer'],
				  'values': ['True', 'False']},
			 'Pollution': {'cpt': [0.9, 0.1], 'parents': [], 'values': ['low', 'high']},
			 'Smoker': {'cpt': [0.3, 0.7], 'parents': [], 'values': ['True', 'False']},
			 'Xray': {'cpt': [0.9, 0.1, 0.2, 0.8],
				  'parents': ['Cancer'],
				  'values': ['positive', 'negative']}})

	def test_V_bn(self):
		self.assertListEqual(self.bn_bn.V,
			['Burglary', 'Earthquake', 'Alarm', 'JohnCalls', 'MaryCalls'])
	
	def test_E_bn(self):
		self.assertDictEqual(self.bn_bn.E,
			{'Alarm': ['JohnCalls', 'MaryCalls'],
				 'Burglary': ['Alarm'],
				 'Earthquake': ['Alarm'],
				 'JohnCalls': [],
				 'MaryCalls': []})

	def test_F_bn(self):
		self.assertDictEqual(self.bn_bn.F,
			{'Alarm': {'cpt': [0.999, 0.001, 0.71, 0.29, 0.06, 0.94, 0.05, 0.95],
			  'parents': ['Earthquake', 'Burglary'],
			  'values': ['No', 'Yes']},
			 'Burglary': {'cpt': [0.999, 0.001], 'parents': [], 'values': ['No', 'Yes']},
			 'Earthquake': {'cpt': [0.998, 0.002], 'parents': [], 'values': ['No', 'Yes']},
			 'JohnCalls': {'cpt': [0.95, 0.05, 0.1, 0.9],
			  'parents': ['Alarm'],
			  'values': ['No', 'Yes']},
			 'MaryCalls': {'cpt': [0.99, 0.01, 0.3, 0.7],
			  'parents': ['Alarm'],
			  'values': ['No', 'Yes']}})

	def test_nodes(self):
		n = list(self.bn_bn.nodes())
		self.assertListEqual(n,
		['Burglary', 'Earthquake', 'Alarm', 'JohnCalls', 'MaryCalls'])

	def test_cpt(self):
		cpt = list(self.bn_bn.cpt('Alarm'))
		self.assertListEqual(cpt,
			[0.999, 0.001, 0.71, 0.29, 0.06, 0.94, 0.05, 0.95])

	def test_card(self):
		self.assertEqual(self.bn_bn.card('Alarm'),2)

	def test_scope(self):
		self.assertListEqual(self.bn_bn.scope('Alarm'),
			['Alarm', 'Earthquake', 'Burglary'])
	
	def test_parents(self):
		self.assertListEqual(self.bn_bn.parents('Alarm'),
			['Earthquake','Burglary'])

	def test_values(self):
		self.assertListEqual(self.bn_bn.values('Alarm'),['No','Yes'])

	def test_values_idx(self):
		self.assertEqual(self.bn_bn.values('Alarm')[1],'Yes')

if __name__ == '__main__':
	unittest.main(exit=False)







# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module provides tools for creating and using an individual factorized representation of a node. See description of factorized representations in :doc:`tablecpdfactorization`.

'''

import sys 

class TableCPDFactor(object):
    '''
    This class represents a factorized representation of a conditional probability distribution table. It contains the attributes *inputvertex*, *inputbn*, *vals*, *scope*, *stride*, and *card*, and the methods *multiplyfactor*, *sumout*, *reducefactor*, and *copy*. 

    '''

    def __init__(self, vertex, bn):
        '''
        This class is constructed with a :doc:`DiscreteBayesianNetwork <discretebayesiannetwork>` instance and a *vertex* name as arguments. First it stores these inputs in *inputvertex* and *inputbn*. Then, it creates a factorized representation of *vertex*, storing the values in *vals*, the names of the variables involved in *scope* the cardinality of each of these variables in *card* and the stride of each of these variables in *stride*.
        
        '''
        self.inputvertex = vertex
        '''The name of the vertex.'''
        self.inputbn = bn
        '''The :doc:`DiscreteBayesianNetwork <discretebayesiannetwork>` instance that the vertex lives in.'''
        
        result = dict( vals = [], stride = dict(), card = [], scope = [])
        root = bn.Vdata[vertex]["cprob"]
        
        # add values
        def explore(_dict, key, depth, totaldepth):
            if depth == totaldepth:
                for x in _dict[str(key)]:
                    result["vals"].append(x)
                return
            else:
                for val in bn.Vdata[bn.Vdata[vertex]["parents"][depth]]["vals"]:
                    ckey = key[:]
                    ckey.append(str(val))
                    explore(_dict, ckey, depth+1, totaldepth)
                    
        if not bn.Vdata[vertex]["parents"]:
            result["vals"] = bn.Vdata[vertex]["cprob"]
        else: 
            td = len(bn.Vdata[vertex]["parents"])
            explore(root, [], 0, td)
        
        # add cardinalities
        result["card"].append(bn.Vdata[vertex]["numoutcomes"])
        if (bn.Vdata[vertex]["parents"] != None):
            for parent in reversed(bn.Vdata[vertex]["parents"]):
                result["card"].append(bn.Vdata[parent]["numoutcomes"])
            
        # add scope
        result["scope"].append(vertex)
        if (bn.Vdata[vertex]["parents"] != None):
            for parent in reversed(bn.Vdata[vertex]["parents"]):
                result["scope"].append(parent)
        
        
        # add strides
        stride = 1
        result["stride"] = dict()
        for x in range(len(result["scope"])):
            result["stride"][result["scope"][x]] = (stride)
            stride *= bn.Vdata[result["scope"][x]]["numoutcomes"]
        
        self.vals = result["vals"]
        '''A flat array of all the values from the CPD.'''
        self.scope = result["scope"]
        '''An array of vertices that affect the vals found in *vals*. Normally, this is the node itself and its parents.'''
        self.card = result["card"]
        '''A list of the cardinalities of each vertex in scope, where cardinality is the number of values that the vertex may take. The cardinalities are indexed according to the vertex's index in *scope*.'''
        self.stride = result["stride"]
        '''A dict of {vertex: value} pairs for each vertex in *scope*, where vertex is the name of the vertex and value is the stride of that vertex in the *vals* array.'''

    def multiplyfactor(self, other):  # cf. PGM 359 
        '''
        Multiply the factor by another :doc:`TableCPDFactor <tablecpdfactor>`. Multiplying factors means taking the union of the scopes, and for each combination of variables in the scope, multiplying together the probabilities from each factor that that combination will be found.
        
        Arguments:
            1. *other* -- An instance of the :doc:`TableCPDFactor <tablecpdfactor>` class representing the factor to multiply by.
                 
        Attributes modified: 
            *vals*, *scope*, *stride*, *card* -- Modified to reflect the data of the new product factor.
                                                         
        For more information cf. Koller et al. 359.

        '''
        if (not isinstance(other, TableCPDFactor)):
            msg = "Error: in method 'multiplyfactor', input was not a TableCPDFactor instance"
            sys.exit(msg)
        j = 0
        k = 0
        result = dict()
        
        # merge scopes
        result["scope"] = self.scope
        result["card"] = self.card
        for x in range(len(other.scope)):
            try:
                result["scope"].index(other.scope[x])
            except: 
                result["scope"].append(other.scope[x])
                result["card"].append(other.card[x])
    
        # calculate possible combinations of scope variables
        possiblevals = 1
        for val in result["card"]:
            possiblevals *= val
        
        # algorithm (see book)
        assignment = [0 for l in range(len(result["scope"]))]
        result["vals"] = []
        for _ in range(possiblevals):
            result["vals"].append(self.vals[j] * other.vals[k])
            for l in range(len(result["scope"])):
                assignment[l] = assignment[l] + 1
                if (assignment[l] == result["card"][l]):
                    assignment[l] = 0
                    try:
                        j = j - (result["card"][l] - 1) * self.stride[result["scope"][l]]
                    except:
                        pass
                    try:
                        k = k - (result["card"][l] - 1) * other.stride[result["scope"][l]]
                    except: 
                        pass
                else:
                    try: 
                        j = j + self.stride[result["scope"][l]]
                    except: 
                        pass
                    try: 
                        k = k + other.stride[result["scope"][l]]
                    except:
                        pass
                    break
            
        # add strides
        stride = 1 
        result["stride"] = dict()
        for x in range(len(result["scope"])):
            result["stride"][result["scope"][x]] = (stride)
            stride *= result["card"][x]
    
        self.vals = result["vals"]
        self.scope = result["scope"]
        self.card = result["card"]
        self.stride = result["stride"]
        
    def sumout(self, vertex):
        '''
        Sum out the variable specified by *vertex* from the factor. Summing out means summing all sets of entries together where *vertex* is the only variable changing in the set. Then *vertex* is removed from the scope of the factor.
        
        Arguments:
            1. *vertex* -- The name of the variable to be summed out.
        
        Attributes modified: 
            *vals*, *scope*, *stride*, *card* -- Modified to reflect the data of the summed-out product factor.
        
        For more information see Koller et al. 297.

        '''
        if (self.scope.count(vertex) == 0):
            msg = "Error: in method 'sumout', vertex '%s' not in scope of factor" % (vertex)
            sys.exit(msg)
        vscope = self.scope.index(vertex)
        vstride = self.stride[vertex]
        vcard = self.card[vscope]
        result = [0 for i in range(len(self.vals)/self.card[vscope])]
        
        # machinery that calculates values in summed out factor
        k = 0
        lcardproduct = 1
        for i in range(vscope):
            lcardproduct *= self.card[i]
        for i in range(len(result)):
            for h in range(vcard):
                result[i] += self.vals[k + (vstride * h)]
            k += 1
            if (k % lcardproduct == 0):
                k += (lcardproduct * (vcard - 1))
        self.vals = result
        
        # modify scope, card, and stride in new factor
        self.scope.remove(vertex)
        del(self.card[vscope])
        for i in range(vscope, len(self.stride)-1):
            self.stride[self.scope[i]] /= vcard
        del(self.stride[vertex])
        
    def reducefactor(self, vertex, value):
        '''
        Reduce the factor knowing that *vertex* equals *value*. Reducing the factor means erasing all possibilities for *vertex* other than *value* from the distribution, and removing *vertex* from the scope.
        
        Arguments:
            1. *vertex* -- The UUID of the variable whose outcome is known.
            2. *value* -- The known outcome of that variable.
        
        Attributes modified: 
            *vals*, *scope*, *stride*, *card* -- Modified to reflect the data of the reduced factor.

        '''
        vscope = self.scope.index(vertex)
        vstride = self.stride[vertex]
        vcard = self.card[vscope]
        result = [0 for i in range(len(self.vals)/self.card[vscope])]
        
        # added step: find value index from evidence
        try:
            index = self.inputbn.Vdata[vertex]['vals'].index(value)
        except:
            raise (Exception, "Second arg was not a possible value of first arg.")
     
        # machinery that calculates values in summed out factor
        k = 0
        lcardproduct = 1
        for i in range(vscope):
            lcardproduct *= self.card[i]
        for i in range(len(result)):
            result[i] += self.vals[k + (vstride * index)]
            k += 1
            if (k % lcardproduct == 0):
                k += (lcardproduct * (vcard - 1))
        self.vals = result
        
        # modify scope, card, and stride in new factor
        self.scope.remove(vertex)
        del(self.card[vscope])
        for i in range(vscope, len(self.stride)-1):
            self.stride[self.scope[i]] /= vcard
        del(self.stride[vertex])
    
    def copy(self):
        '''Return a copy of the factor.'''
        copy = TableCPDFactor(self.inputvertex, self.inputbn)
        copy.vals = self.vals[:]
        copy.stride = self.stride.copy()
        copy.scope = self.scope[:]
        copy.card = self.card[:]
        return copy
    
        
    



# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
Nearly all of the functions of this library require key indexing, which means it deals with dictionaries internally. This module deals with loading dictionaries and handles automatically converting from python-style dictionaries to condensed (no excess white space) JSON-style dictionaries.

'''
import sys
import json
import string

class Dictionary(object):
    '''
    This class represents a JSON-style, key-indexable dictionary of data. It contains the attribute *alldata* and the method *dictload*. 
    '''

    def __init__(self):
        self.alldata = None
        '''An internal representation of a key-indexable dictionary.'''

    def dictload(self, path):
        '''
        Load a dictionary from a JSON-like text in a text file located at *path* into the attribute *alldata*.
        
        In order for this function to execute successfully, the text file must have the proper formatting, particularly with regard to quotation marks. See :doc:`unittestdict` for an example. Specifically, the function can get rid of excess whitespace, convert ``.x`` to ``0.x`` in decimals, and convert ``None`` to ``null``, but nothing else.

        Arguments:
            
            1. *path* -- Path to the text file (e.g. "mydictionary.txt")
        
        Attributes modified: 
        
            1. *alldata* -- The entire loaded dictionary.
        
        The function also returns an error if nothing was loaded into *alldata*.

        '''
        f = open(path, 'r')
        ftext = f.read() 
        assert (ftext and isinstance(ftext, str)), "Input file is empty or could not be read."


        # alter for json input, if necessary
        loaded = False
        try:
            self.alldata = json.loads(ftext)
            loaded = True
        except ValueError:
            pass

        if not loaded:
            try: 
                ftext = ftext.translate(None, '\t\n ')
                ftext = ftext.replace(':', ': ')
                ftext = ftext.replace(',', ', ')
                ftext = ftext.replace('None', 'null')
                ftext = ftext.replace('.', '0.')
                self.alldata = json.loads(ftext)
            except ValueError:
                raise (ValueError, "Convert to JSON from input file failed. Check formatting.")
        f.close()
        
        assert isinstance(self.alldata, dict), "In method dictload, path did not direct to a proper text file."
        
        
        



"""
******************
CliqueTree Class 
&
Clique Class
******************

This is a class for creating/manipulating Junction (Clique) Trees,
and performing inference over them. The advantage of clique trees
over traditional variable elimination over the original Bayesian 
network is that clique trees allow you to compute marginal
probabilities of MULTIPLE variables without having to run the
entire algorithm over. 

Therefore, if you have to query the Bayesian network many times 
-- and you want the exact marginal values -- then it might be 
best to use the clique tree data structure for inference.
If you need to compute the marginal distribution for all variables,
but you dont mind an approximate, a sampling algorithm is probably
the way to go.

In general, the junction tree algorithms generalize 
Variable Elimination to the efficient, simultaneous execution 
of a large class of queries.

NOTE: A cluster graph is a generalization of the clique tree
data structure - to generate a clique tree, you first generate
a cluster graph, then simply calculate a maximum spanning tree.
In other words, a clique tree can be considered as a special
type of cluster graph.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import networkx as nx
from copy import copy

from bamt.external.pyBN.classes.factorization import Factorization

from bamt.external.pyBN.utils.graph import *



class CliqueTree(object):
    """
    CliqueTree Class

    Let G be a chordal graph (ie. a graph such that no cycle includes more
    than three nodes), then a CliqueTree is a tree H such that each
    maximal clique C in G is a node in H.

    Attributes
    ----------
    - bn 
        - BayesNet object

    - V
        - Vertices -> list

    - E
        - Edges -> dictionary

    - C
        - Cliques -> a dictionary where key = vertex idx,
                value = Clique object

    """

    def __init__(self, bn):
        """
        Instantiate a CliqueTree object.

        Arguments
        ---------
        *bn*: a BayesNet object

        Notes
        -----
        Ideally, the Factor class should be used as the
        cliques instead of the Clique class (because it's
        just a watered down version of the Factor class)        
        
        """
        ####
        self.V = None
        self.E = None
        self.C = None
        ###

        self.bn = bn
        self._F = Factorization(bn)
        self.initialize_tree()

        

    #def __repr__(self):
       # return self.C

    def __iter__(self):
        for clique in self.C.values():
            yield clique

    def __getitem__(self, rv):
        """
        Returns Clique of passed-in rv
        """
        return self.C[rv]

    def parents(self, v):
        p = []
        for rv in self.V:
            if v in self.E[rv]:
                p.append(v)
        return p

    def children(self, n):
        return self.E[n]

    def dfs_postorder(self, root):
        G = nx.Graph(self.E)
        tree_graph = nx.dfs_tree(G,root)
        clique_ordering = list(nx.dfs_postorder_nodes(tree_graph,root))
        return clique_ordering

    def initialize_tree(self):
        """
        Initialize the structure of a clique tree, using
        the following steps:
            - Moralize graph (i.e. marry parents)
            - Triangulate graph (i.e. make graph chordal)
            - Get max cliques (i.e. community/clique detection)
            - Max spanning tree over sepset cardinality (i.e. create tree)
        
        """
        ### MORALIZE GRAPH & MAKE IT CHORDAL ###
        chordal_G = make_chordal(self.bn) # must return a networkx object
        V = chordal_G.nodes()

        ### GET MAX CLIQUES FROM CHORDAL GRAPH ###
        C = {} # key = vertex, value = clique object
        max_cliques = reversed(list(nx.chordal_graph_cliques(chordal_G)))
        for v_idx,clique in enumerate(max_cliques):
            C[v_idx] = Clique(set(clique))

        ### MAXIMUM SPANNING TREE OVER COMPLETE GRAPH TO MAKE A TREE ###
        weighted_edge_dict = dict([(c_idx,{}) for c_idx in range(len(C))])
        for i in range(len(C)):
            for j in range(len(C)):
                if i!=j:
                    intersect_cardinality = len(C[i].sepset(C[j]))
                    weighted_edge_dict[i][j] = -1*intersect_cardinality
        mst_G = mst(weighted_edge_dict)
        ### SET V,E,C ###
        self.E = mst_G # dictionary
        self.V = mst_G.keys() # list
        self.C = C

        
        ### ASSIGN EACH FACTOR TO ONE CLIQUE ONLY ###
        v_a = dict([(rv, False) for rv in self.bn.nodes()])
        for clique in self.C.values():
            temp_scope = []
            for var in v_a:
                if v_a[var] == False and set(self.bn.scope(var)).issubset(clique.scope):
                    temp_scope.append(var)
                    v_a[var] = True
            clique._F = Factorization(self.bn, temp_scope)

        ### COMPUTE INITIAL POTENTIAL FOR EACH FACTOR ###
        # - i.e. multiply all of its assigned factors together
        for i, clique in self.C.items():
            if len(self.parents(i)) == 0:
                clique.is_ready = True
            clique.initialize_psi()


class Clique(object):
    """
    Clique Class

    *scope* : a set of variables in the clique's scope

    *_f* : a factorization object that contains only the
        factors of variables in the clique's scope

    *psi* : a factor
        The potential of the clique.

    *belief* : a factor
        The factor which holds the marginal/conditional
        probabilities of the relevant nodes after 
        belief propagation, etc

    *messages_received* : a list of Factors
        The messages the clique has received from its neighbors -
        stored so that the beliefs can be calculated after
        belief propagation, etc.

    *is_ready* : a boolean
        Whether the clique is ready to send a message -> must
        have RECEIVED messages from all of its neighbors first.

    """

    def __init__(self, scope):
        """
        Instantiate a clique object

        Arguments
        ---------
        *scope* : a python set
            The set of variables in the cliques scope,
            i.e. the main var and its parents


        """
        self.scope = scope
        self._F = None
        
        self.psi = None # Psi should never change -> Factor object
        self.belief = None
        
        self.messages_received = []
        self.is_ready = False

    def __repr__(self):
        return str(self.scope)

    def __rshift__(self, other_clique):
        """
        Send a message from self to other_clique
        """
        self.send_message(other_clique)

    def __lshift__(self, other_clique):
        """
        Send a message from other_clique to self
        """
        other_clique.send_message(self)

    def send_message(self, parent):
        """
        Send a message to the parent clique.

        To send a message from X to Y, you
        must first take the potential (self.psi) of
        X and sum out the variables NOT in the sepset
        of Y.

        THEN, if X has received any messages, the
        potential (self.psi) must be multiplied by
        all of the received messaged.

        Arguments
        ---------
        *parent* : a string
            The parent to which the message will
            be sent.

        """


        # First generate Belief = Original_Psi * all received messages
        if len(self.messages_received) > 0:
            # if there are messages received, mutliply them in to psi first
            if not self.belief:
                self.belief = copy(self.psi)
            for msg in self.messages_received:
                self.belief *= msg
            #self.belief.merge_multiply(self.messages_received)
        else:
            # if there are no messages received, simply move on with psi
            if not self.belief:
                self.belief = copy(self.psi)
        # generate message as belief with Ci - Sij vars summed out
        #vars_to_sumout = list(self.scope.difference(self.sepset(parent)))
        msg_to_send = copy(self.belief)
        for var_to_sumout in self.scope.difference(self.sepset(parent)):
            msg_to_send /= var_to_sumout
        #message_to_send = copy(self.belief)
        #message_to_send.sumout_var_list(vars_to_sumout)
        parent.messages_received.append(msg_to_send)

    def initialize_psi(self):
        """
        Compute a new psi (cpt) in order to 
        set the clique's belief. This involves
        multiplying the factors in the Clique together.
        """
        assert (len(self._F) != 0), 'No Factors assigned to this clique.'

        if len(self._F) == 1:
            self.psi = copy(self._F[0])
            self.belief = copy(self.psi)
        else:
            self.psi = max(self._F, key=lambda x: len(x.cpt))
            for f in self._F:
                self.psi *= f
            self.belief = copy(self.psi)

    def send_initial_message(self, other_clique):
        """
        NOT SURE IF THIS IS NEEDED.

        Send the first message to another clique.

        Arguments
        ---------
        *other_clique* : a different Clique object

        """
        psi_copy = copy(self.psi)
        sepset = self.sepset(other_clique)
        sumout_vars = self.scope.difference(sepset)
        # sum out variables not in the sepset of other_clique
        psi_copy.sumout_var_list(list(sumout_vars))


        #psi_copy.cpt = psi_copy.cpt.loc[:,[c for c in psi_copy.cpt.columns if 'Prob' not in c]]
        #psi_copy.cpt[str('Prob-Val-' + str(np.random.randint(0,1000000)))] = 1
        print('Init Msg: \n', psi_copy.cpt)
        other_clique.messages_received.append(psi_copy)

        self.belief = copy(self.psi)

    def collect_beliefs(self):
        """
        Collecting beliefs is done by taking the
        potential (self.psi) of X and multiplying by all of
        the messages which X has received.

        NOTE: once beliefs are collected, the messages
        which the Clique has received are cleared out.

        Notes
        -----
        A root node (i.e. one that doesn't send a message) must run collect_beliefs
        since they are only collected at send_message()

        Also, we collect beliefs at the end of loopy belief propagation (approx. inference)
        since the main algorithm is just sending messages for a while.
        """
        if len(self.messages_received) > 0:
            self.belief = copy(self.psi)
            for msg in self.messages_received:
                self.belief *= msg
            self.belief 
            self.messages_received = []
        else:
            self.belief = copy(self.belief)

    def sepset(self, other_clique):
        """
        The sepset of two cliques is the set of
        variables in the intersection of the two
        cliques' scopes.

        Arguments
        ---------
        *other_clique* : a Clique object

        """
        return self.scope.intersection(other_clique.scope)

    def marginalize_over(self, target):
        """
        Marginalize the cpt (belief) over a target variable.

        Arguments
        ---------
        *target* : a string
            The target random variable.

        """
        blf = copy(self.belief)
        blf.sumover_var(target)
        return blf


















"""
**************************
Discretize Continuous Data
**************************

Since pyBN only handles Discrete Bayesian Networks,
and therefore only handles discrete data, it is
important to have effective functions for 
discretizing continuous data. This code aims to
meet that goal.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np
from copy import copy
import random
import pandas as pd



"""
>>dediscretize
Function to convert labels back to continuous values

>>data_with_empty_column
Helper function to fill with gaps in a desired column in data


"""









def discretize(data, cols=None, bins=None):
	"""
	Discretize the passed-in dataset. These
	functions will rely on numpy and scipy
	for speed and accuracy... no need to
	reinvent the wheel here. Therefore, pyBN's
	discretization methods are basically just
	wrappers for existing methods.

	The bin number defaults to FIVE (5) for
	all columns if not passed in.

	Arguments
	---------
	*data* : a nested numpy array

	*cols* : a list of integers (optional)
		Which columns to discretize .. defaults
		to ALL columns

	*bins* : a list of integers (optional)
		The number of bins into which each column
		array will be split .. defaults to 5 for
		all columns

	Returns
	-------
	*data* : a discretized copy of original data

    *dic_list* : A dictionary with the value of the middle of each interval
    key = number of the interval
    value = the middle of the interval

    *steps*:  A list with step value of each column in *cols*

	Effects
	-------
	None

	Notes
	-----
	- Should probably add more methods of discretization
		based on mean/median/mode, etc.
	"""
	if bins is not None:
		assert (isinstance(bins, list)), 'bins argument must be a list'
	else:
		try:
			bins = [5]*data.shape[1]
		except ValueError:
			bins = [5]
	
	if cols is not None:
		assert (isinstance(cols,list)), 'cols argument must be a list'
	else:	
		try:
			cols = range(data.shape[1])
		except ValueError:
			cols = [0]

	data = copy(data)
	dic_list = []
	steps = []

	minmax = list(zip(np.amin(data,axis=0),np.amax(data,axis=0)))
	for i, c in enumerate(cols):
		# get min and max of each column
		_min, _max = minmax[c]
		# create the bins from np.linspace
		_bins, step = np.linspace(_min,_max,bins[i], retstep=True)
		# discretize with np.digitize(col,bins)
		data[:,c] = np.digitize(data[:,c],_bins)
		steps.append(int(step))
		dic = {}
		for j in range(1, bins[i]+1, 1):
			dic[j] = (_bins[j-1]) + int(step/2)
		dic_list.append(dic)
	return np.array(data,dtype=np.int32,copy=False), dic_list, steps



def dediscretize(data, dic_labels, steps, cols):
	
	data = copy(data)
	for i, c in enumerate(cols):
		for j in range(data.shape[0]):
			data[j, c] = dic_labels[i][data[j, c]] + random.randint(0,int((steps[i])/2)) - random.randint(0,int((steps[i])/2))
	return pd.DataFrame(data)

def data_with_empty_column(data, index_list):
    empty_data = []
    for e in data.values:
        empty_data.append(list(e.astype('str')))
    for i in range(len(empty_data)):
        for j in index_list:
            empty_data[i][j] = None
    return empty_data












"""
***************
UnitTest
Marginal Approx
***************

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from bamt.external.pyBN.inference.marginal_approx import marginal_fs_a, marginal_lws_a, marginal_gs_a
from bamt.external.pyBN.readwrite.read import read_bn

class MarginalApproxTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn = read_bn(os.path.join(self.dpath,'cmu.bn'))

	def tearDown(self):
		pass

	def test_forward_sample(self):
		np.random.seed(3636)
		self.assertDictEqual(marginal_fs_a(self.bn,n=1000),
			{'Alarm': {'No': 0.996, 'Yes': 0.004},
			 'Burglary': {'No': 0.998, 'Yes': 0.002},
			 'Earthquake': {'No': 1.0, 'Yes': 0.0},
			 'JohnCalls': {'No': 0.946, 'Yes': 0.054},
			 'MaryCalls': {'No': 0.99, 'Yes': 0.01}}
			 )

	def test_likelihood_weighted_sample(self):
		"""
		THIS HAS BEEN VALIDATED AGAINST BNLEARN
		"""
		np.random.seed(3636)
		self.assertDictEqual(marginal_lws_a(self.bn,evidence={'Burglary':'Yes'}),
			{'Alarm': {'No': 0.058, 'Yes': 0.942},
			 'Burglary': {'No': 0.0, 'Yes': 1.0},
			 'Earthquake': {'No': 0.997, 'Yes': 0.003},
			 'JohnCalls': {'No': 0.136, 'Yes': 0.864},
			 'MaryCalls': {'No': 0.343, 'Yes': 0.657}})

	def test_lws_allevidence(self):
		self.assertDictEqual(marginal_lws_a(self.bn,evidence={'Burglary':'Yes','Alarm':'Yes',
			'Earthquake':'Yes','JohnCalls':'Yes','MaryCalls':'Yes'}),
			{'Alarm': {'No': 0.0, 'Yes': 1.0},
			 'Burglary': {'No': 0.0, 'Yes': 1.0},
			 'Earthquake': {'No': 0.0, 'Yes': 1.0},
			 'JohnCalls': {'No': 0.0, 'Yes': 1.0},
			 'MaryCalls': {'No': 0.0, 'Yes': 1.0}})

	def test_gibbs(self):
		np.random.seed(3636)
		self.assertDictEqual(marginal_gs_a(self.bn,n=1000,burn=200),
			{'Alarm': {'No': 0.9988, 'Yes': 0.0},
				 'Burglary': {'No': 0.9988, 'Yes': 0.0},
				 'Earthquake': {'No': 0.9975, 'Yes': 0.0013},
				 'JohnCalls': {'No': 0.9313, 'Yes': 0.0675},
				 'MaryCalls': {'No': 0.9838, 'Yes': 0.015}})













"""
*************
Factorization 
Class
*************
"""
from bamt.external.pyBN.classes.factor import Factor

from collections import OrderedDict
from copy import copy,deepcopy
import numpy as np

class Factorization(object):

	def __init__(self, bn, nodes=None):
		"""
		Initialize a Factorization object.

		The *nodes* argument allows you to make a Factorization object
		that includes only a subset of the random variables in the
		passed-in BayesNet object. This is mostly useful for CliqueTree
		algorithms.
		
		"""
		self.bn = bn

		if nodes is not None:
			self._phi = [Factor(bn,rv) for rv in nodes]
		else:
			self._phi = [Factor(bn,rv) for rv in bn.nodes()]

		## MAP-BASED ATTRIBUTES ##
		self.map_factors = OrderedDict()
		self.map_assignment = dict()
		self.map_prob = -1

	def refresh(self):
		"""
		Refresh Factorization attributes.
		"""
		self._phi = [Factor(self.bn,rv) for rv in self.bn.nodes()]
		self.map_factors = OrderedDict()
		self.map_assignment = dict()
		self.map_prob = -1

	def __getitem__(self, idx):
		return self._phi[idx]

	def __iter__(self):
		for phi in self._phi:
			yield phi

	def __len__(self):
		return len(self._phi)

	def __div__(self, rv):
		"""
		Overloads division operator for
		Sum-Product Variable Elimination pass
		over passed-in var.
		"""
		self.sum_product_eliminate_var(rv)
		return self

	def __floordiv__(self, rv):
		"""
		Overloads floor division operator for
		Max-Product Variable Elimination pass
		over passed-in var.
		"""
		self.max_product_eliminate_var(rv)
		return self

	def __sub__(self, rv_val):
		"""
		Overloads subtract operator for
		reducing EVERY relevant factor
		by passed-in rv_val
		"""
		self.map_assignment[rv_val[0]] = rv_val[1] # append
		for phi in self._phi:
			if rv_val[0] in phi.scope:
				phi -= (rv_val[0],rv_val[1]) # reduce by evidence
		return self

	def sum_product_eliminate_var(self, rv):
		relevant_factors = self.relevant_factors(rv)
		irrelevant_factors = self.irrelevant_factors(rv)

		# mutliply all relevant factors
		psi = relevant_factors[0]
		for i in range(1,len(relevant_factors)):
			psi *= relevant_factors[i]

		# Take Sum over psi for rv
		psi /= rv # sumout
		irrelevant_factors.append(psi) # add sum-prod factor back in

		self._phi = irrelevant_factors

	def max_product_eliminate_var(self, rv):
		relevant_factors = self.relevant_factors(rv)
		irrelevant_factors = self.irrelevant_factors(rv)

		# mutliply all relevant factors
		psi = relevant_factors[0]
		for i in range(1,len(relevant_factors)):
			psi *= relevant_factors[i]

		self.map_factors[rv] = deepcopy(psi)
		# Take Max over psi for rv
		psi //= rv # maxout
		irrelevant_factors.append(psi) # add sum-prod factor back in

		self._phi = irrelevant_factors

	def traceback_map(self):
		nodes = self.map_factors.keys()
		idx = len(self.map_factors)
		for rv in reversed(self.map_factors.keys()):
			f = self.map_factors[rv]
			# reduce factor by variables upstream
			for i in range(idx,len(self.map_factors)):
				if nodes[i] in f.scope:
					f -= (nodes[i],self.map_assignment[nodes[i]])
			# choose maximizing assignment conditioned on upstream vars
			self.map_assignment[rv] = self.bn.values(rv)[np.argmax(self.map_factors[rv].cpt)]
			idx-=1
		return self.map_assignment

	def consolidate(self):
		final_phi = self._phi[0]
		for i in range(1,len(self._phi)):
			final_phi *= _phi[i]
		#final_phi.normalize()
		return final_phi

	def relevant_factors(self, rv):
		return [f for f in self._phi if rv in f.scope]

	def irrelevant_factors(self, rv):
		return [f for f in self._phi if rv not in f.scope]













# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module facilitates the process of creating ordered graph skeletons by topologically ordering them automatically.

'''

from bamt.external.libpgm.graphskeleton import GraphSkeleton


class OrderedSkeleton(GraphSkeleton):
    '''
    This class represents a graph skeleton (see :doc:`graphskeleton`) that is always topologically ordered.
    
    '''
    
    def __init__(self, graphskeleton=None):
        self.V = None
        '''A list of names of vertices'''
        self.E = None
        '''A list of [origin, destination] pairs of verties that constitute edges.'''

    def load(self, path):
        '''Loads a dictionary from a file located at *path* in the same manner as :doc:`graphskeleton`, but includes a step where it topologically orders the nodes.'''
    
        self.dictload(path)
        self.V = self.alldata["V"]
        self.E = self.alldata["E"]

        # topologically order
        self.toporder()

        # free unused memory
        del self.alldata



import os,sys,inspect
# currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# parentdir = os.path.dirname(currentdir)
# sys.path.insert(0,parentdir)
from copy import copy
import math
from typing import List
import numpy as np
import pandas as pd


def nodes_from_edges(edges: list):
    """
    Retrieves all nodes from the list of edges.
            Arguments
    ----------
    *edges* : list

    Returns
    -------
    *nodes* : list

    Effects
    -------
    None
    """
    all_nodes = []
    for e in edges:
        all_nodes.extend(e)
    return list(set(all_nodes))

def edges_to_dict(edges: list):
    """
    Transfers the list of edges to the dictionary of parents.
            Arguments
    ----------
    *edges* : list

    Returns
    -------
    *parents_dict* : dict

    Effects
    -------
    None
    """
    nodes = nodes_from_edges(edges)
    parents_dict = {var: [] for var in nodes}
    for e in edges:
        parents_dict[e[1]].append(e[0])
    return parents_dict